#include <ctype.h>
#include <stdbool.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <stralloc.h>
#include <afblib/inbuf.h>
#include <afblib/inbuf_scan.h>
#include <afblib/outbuf.h>
#include <afblib/outbuf_printf.h>

#define MAX_COLOURS 9 /* just digits, please */
#define MAX_LENGTH 20

typedef struct {
   unsigned int len; /* length of the code */
   unsigned int colours; /* number of different colours
      (or digits) used in the code */
   bool unique; /* are the colours (or digits) unique in the code? */
   unsigned int guesses; /* number of guesses so far */
   bool running; /* has the game been initialized and not yet guessed? */
   char code[MAX_LENGTH + 1]; /* actual code */
   unsigned int colourset; /* bitset of all colours used in the code */
   unsigned int count[MAX_COLOURS]; /* number of occurrences of each colour */
} Game;

/* create a new code for a new game using the
   parameters len, colours, and unique */
void init_game(Game* game) {
   for (unsigned int i = 0; i < MAX_COLOURS; ++i) {
      game->count[i] = 0;
   }
   unsigned int set = 0; /* bitset of colours */
   for (unsigned int i = 0; i < game->len; ++i) {
      unsigned int colour; unsigned int mask;
      do {
	 colour = rand() % game->colours;
	 mask = (1 << colour);
      } while (game->unique && (set & mask));
      set |= mask;
      ++game->count[colour];
      game->code[i] = colour + '0' + 1;
   }
   game->code[game->len] = 0;
   game->guesses = 0;
   game->colourset = set;
   game->running = true;
}

/* macros used for bitset operations for colours */
#define COLOUR(colour) \
   ((colour)-'0'-1)
#define IN_SET(colour, set) \
   ((1<<COLOUR(colour)) & set)

/* evaluate a guess by setting *black_counter and *white_counter,
   if the guess appears irregular, false is returned */
bool evaluate_guess(Game* game, stralloc* guess,
      unsigned int* black_counter, unsigned int* white_counter) {
   unsigned black = 0; unsigned int white = 0;
   if (game->unique) {
      for (unsigned int i = 0; i < game->len; ++i) {
	 if (guess->s[i] == game->code[i]) {
	    ++black;
	 } else if (guess->s[i] == '0' ||
	       COLOUR(guess->s[i]) >= game->colours) {
	    /* colour used which cannot appear in the code */
	    return false;
	 } else if (IN_SET(guess->s[i], game->colourset)) {
	    ++white;
	 }
      }
   } else {
      unsigned int count[MAX_COLOURS] = {0};
      for (unsigned int i = 0; i < game->len; ++i) {
	 if (guess->s[i] == '0' || COLOUR(guess->s[i]) >= game->colours) {
	    /* colour used which cannot appear in the code */
	    return false;
	 }
	 count[COLOUR(guess->s[i])]++;
	 if (guess->s[i] == game->code[i]) {
	    ++black;
	 }
      }
      for (unsigned int i = 0; i < game->colours; ++i) {
	 if (count[i] <= game->count[i]) {
	    white += count[i];
	 } else {
	    white += game->count[i];
	 }
      }
      white -= black;
   }
   ++game->guesses;
   *black_counter = black;
   *white_counter = white;
   return true;
}

/* convert decimal sequence of digits to an integer value */
unsigned int get_value(stralloc* param) {
   unsigned int val = 0;
   for (unsigned int i = 0; i < param->len; ++i) {
      val = val * 10 + param->s[i] - '0';
   }
   return val;
}

char* help[] = {
   "Either a code using digits 1 to 9 for up to nine colours",
   "or a command may be given:",
   "",
   "colours n       set number of different colours",
   "game            start a new game",
   "help            print this information",
   "len n           set code length",
   "quit            terminate session",
   "unique          toggle uniqueness",
};

void print_status(outbuf* out, Game* game) {
   outbuf_printf(out, "S len = %u, colours = %u, unique = %s\r\n",
      game->len, game->colours, (game->unique? "true": "false"));
}

/* reveal the code if the game was still running
   and finish the game */
void finish_game(outbuf* out, Game* game) {
   if (game->running && game->guesses > 0) {
      outbuf_printf(out, "C too bad you gave up after %u fruitless attempts"
	 " to guess %.*s\r\n", game->guesses, game->len, game->code);
      game->guesses = 0;
   }
   game->running = false;
}

void mastermind_session(int fd) {
   inbuf in = {fd};
   outbuf out = {fd};
   /* seed random generator here, not im main();
      otherwise all sessions would share the same
      sequence of codes */
   srand(getpid() ^ time(0));
   /* initialize game with default parameters */
   Game game = {.len = 4, .colours = 6, .unique = true};
   outbuf_printf(&out, "S Welcome to MasterMind!\r\n");
   outbuf_flush(&out);
   /* components of a request line */
   stralloc command = {0};
   stralloc parameter = {0};
   for(;;) { /* as long as valid commands are coming... */
      /* read next command */
      int res = inbuf_scan(&in,
	    "(\\w+|\\d+)"             /* command or guess */
	    "(?:[ \t]+(\\d+))?"       /* optional parameter */
	    "\r?\n",
	 &command, &parameter);
      if (res < 0) break; /* end of input */
      if (res == 0) {
	 outbuf_printf(&out, "F invalid command\r\n");
	 outbuf_flush(&out);
	 continue;
      }
      /* the parameter, if given, is always a decimal number */
      unsigned int value = 0;
      if (res == 2) {
	 value = get_value(&parameter);
      }
      if (stralloc_equals(&command, "game") && res == 1) {
	 /* start a new game */
	 finish_game(&out, &game);
	 init_game(&game); print_status(&out, &game);
      } else if (stralloc_equals(&command, "len") && res == 2) {
	 if (value > 0 && value <= MAX_LENGTH) {
	    if (game.unique && value > game.colours) {
	       outbuf_printf(&out, "F length exceeds number of colours\r\n");
	    } else {
	       finish_game(&out, &game);
	       outbuf_printf(&out, "S len = %u\n", value);
	       game.len = value;
	    }
	 } else {
	    outbuf_printf(&out, "F invalid length value\r\n");
	 }
      } else if (stralloc_equals(&command, "colours") && res == 2) {
	 if (value > 1 && value <= MAX_COLOURS) {
	    if (game.unique && value < game.len) {
	       outbuf_printf(&out, "F number of colours must be >= length\r\n");
	    } else {
	       finish_game(&out, &game);
	       outbuf_printf(&out, "S colours = %u\n", value);
	       game.colours = value;
	    }
	 } else {
	    outbuf_printf(&out, "F invalid colours value\r\n");
	 }
      } else if (stralloc_equals(&command, "unique") && res == 1) {
	 if (!game.unique && game.colours < game.len) {
	    outbuf_printf(&out,
	       "F too few colours to switch to uniqueness\r\n");
	 } else {
	    finish_game(&out, &game);
	    game.unique = !game.unique;
	    outbuf_printf(&out, "S unique = %s\n",
	       game.unique? "true": "false");
	 }
      } else if (stralloc_equals(&command, "quit")) {
	 finish_game(&out, &game);
	 break;
      } else if (stralloc_equals(&command, "help")) {
	 for (unsigned int i = 0; i < sizeof(help)/sizeof(help[0]); ++i) {
	    outbuf_printf(&out, "C %s\r\n", help[i]);
	 }
	 print_status(&out, &game);
      } else if (isdigit(command.s[0])) {
	 if (command.len != game.len) {
	    outbuf_printf(&out, "F wrong code length\r\n");
	 } else {
	    if (!game.running) {
	       init_game(&game);
	    }
	    unsigned int black; unsigned int white;
	    if (evaluate_guess(&game, &command, &black, &white)) {
	       if (black == game.len) {
		  outbuf_printf(&out,
		     "C You have solved this code in %u steps!\r\n",
		     game.guesses);
		  game.running = false;
	       }
	       outbuf_printf(&out, "S %.*s Black: %d  White: %d\r\n",
		  command.len, command.s, black, white);
	    } else {
	       outbuf_printf(&out, "F invalid guess\r\n");
	    }
	 }
      } else {
	 outbuf_printf(&out, "F invalid command; try \"help\"\r\n");
      }
      outbuf_flush(&out);
   }
   outbuf_printf(&out, "S bye\r\n");
   outbuf_flush(&out);
}
